from pyxel_engine.event_handler import EventHandler
from pyxel_engine.utils import smallest
from PIL import Image, ImageTk
import keyboard
import _thread
import tkinter as tk
import time
import sys
import traceback
import pathlib
import logging
import os
from math import sqrt, pow
from ctypes import windll

engine_setting = {
    'TPS': 20,
    'TPS_counter': False,
    'catch_error': True,
    'crash_path': 'crash.txt',
    'logging_level': logging.DEBUG,
    'log_path': 'log.txt',
    'resolution': (260, 140)
}

__version__ = '0.0.9'

engineInitialized = False

onEngineStarted = EventHandler()

renderTick = EventHandler()
postProcessing = EventHandler()

drawCollision = EventHandler()
checkCollision = EventHandler()

engineTick = EventHandler()

screen_width = None
screen_height = None
game = None
window = None
game_map = None
tps = 42

current_background = None
background_collision = None

last_tick = time.time_ns() / (10 ** 9)
next_tick = time.time_ns() / (10 ** 9) + 1 / targetedTPS
tick = 0

registry = set()

logging.getLogger('PIL').setLevel(logging.WARNING)
with open(engine_setting['log_path'], 'w'):
    pass

print('Using Pyxel Engine ' + __version__)


class EngineError(Exception):
    pass


class Schedule:
    def __init__(self, timeout, callback, *args):
        global tick
        self.started_tick = tick
        self.callback = callback
        self.args = args
        self.timeout = timeout
        engineTick.add(self.tick)
        logging.debug('Scheduled ' + str(callback) + ' in ' + str(timeout) + ' tick(s)')

    def tick(self):
        global tick
        if self.started_tick != tick:
            self.timeout = self.timeout - 1

        if self.timeout == 0:
            self.callback(*self.args)
            del self


schedule = Schedule


class HitBox:
    def __init__(self, owner, basic_box=None, complex_box=None, debug=False):
        if basic_box is None and complex_box is None:
            raise ValueError('One argument must be provided')

        self.owner = owner
        self.debug = debug

        self.box = []

        if basic_box is not None:
            blank_line = []
            for i in range(0, basic_box[0]):
                blank_line.append(True)
            for i in range(0, basic_box[1]):
                self.box.append(blank_line)
            self.size = basic_box
            self.type = 'basic'

        if complex_box is not None:
            self.box = complex_box
            self.size = (len(complex_box), len(complex_box[0]))
            self.type = 'complex'

        self.onObjectOverlapBegin = EventHandler()
        self.onObjectOverlapEnd = EventHandler()
        self.lastOverlap = set()

        drawCollision.add(self.draw_hitbox)
        checkCollision.add(self.check_collision)

    def draw_hitbox(self):
        global game_map

        if hasattr(self.owner, 'anchor') is True:
            x = self.owner.x + self.owner.anchor[0]
            y = self.owner.y + self.owner.anchor[1]
        else:
            x = self.owner.x
            y = self.owner.y

        if self.type == 'basic':
            for k in range(x, x + self.size[1]):
                for v in range(y, y + self.size[0]):
                    game_map[v][k].add(self.owner)

        if self.type == 'complex':
            for k in range(self.size[1]):
                for v in range(self.size[0]):
                    if self.box[v][k] is True:
                        game_map[v + y][k + x].add(self.owner)

        if self.debug is True:
            logging.debug('Drew hitbox of ' + self.owner)

    def check_collision(self):
        global game_map

        if hasattr(self.owner, 'anchor') is True:
            x = self.owner.x + self.owner.anchor[0]
            y = self.owner.y + self.owner.anchor[1]
        else:
            x = self.owner.x
            y = self.owner.y

        out = set()

        for k in range(x, x + self.size[0]):
            for v in range(y, y + self.size[1]):
                if len(game_map[v][k]) != 0:
                    for item in game_map[v][k]:
                        if item != self.owner:
                            out.add(item)

        if len(out - self.lastOverlap) != 0:
            logging.debug(str(self.owner) + ' overlapped by ' + str(out))
            self.lastOverlap = out
            self.onObjectOverlapBegin.run(other=out)

        if len(self.lastOverlap - out) != 0:
            self.lastOverlap = out
            self.onObjectOverlapEnd.run(other=self.lastOverlap - out)

    def make_projection(self, x, y):
        global game_map

        if hasattr(self.owner, 'anchor') is True:
            x = x + self.owner.anchor[0]
            y = y + self.owner.anchor[1]

        out = set()

        for k in range(x, x + self.size[0]):
            for v in range(y, y + self.size[1]):
                if len(game_map[v][k]) != 0:
                    for item in game_map[v][k]:
                        if item != self.owner:
                            out.add(item)

        if len(out) != 0 and self.debug is True:
            logging.debug('Projection of ' + str(self.owner) + ' at ' + str(x) + ' ' + str(y) + ' failed')

        return out


class Sprite:
    def __init__(self, owner, path=None, image=None, debug=False):
        if path is None and image is None:
            raise ValueError('One argument must be provided')

        self.debug = debug
        self.owner = owner
        renderTick.add(self.render)

        if path is not None:
            self.image = Image.open(path)
        if image is not None:
            self.image = image

    def render(self):
        global game

        if hasattr(self.owner, 'anchor') is True:
            x = self.owner.x + self.owner.anchor[0]
            y = self.owner.y + self.owner.anchor[1]
        else:
            x = self.owner.x
            y = self.owner.y

        if self.debug is True:
            logging.debug(str(self.owner) + '\'s sprite rendered at ' + str(x) + ' ' + str(y))
        game.paste(self.image, (x, y), self.image)


class Background:
    def __init__(self, path=None, image=None, x=0, y=0):
        if path is None and image is None:
            raise ValueError('One argument must be provided')

        if path is not None:
            self.image = Image.open(path)
        if image is not None:
            self.image = image

        self.x = x
        self.y = y


class BackgroundCollision:
    def __init__(self, collision):
        self.collision = collision


keybind = set()


class Keybind:
    def __init__(self, key):
        global keybind

        keybind.add(self)
        self.key = key

        self.on_press = EventHandler()
        self.on_release = EventHandler()

    def change_key(self, key):
        self.key = key

    def delete(self):
        del self


def keyboard_hook_on_press(event):
    global keybind

    for bind in keybind:
        if event.name == bind.key or bind.key == 'global':
            bind.on_press.trigger()


def keyboard_hook_on_release(event):
    global keybind

    for bind in keybind:
        if event.name == bind.key or bind.key == 'global':
            bind.on_release.trigger()


keyboard.on_press(keyboard_hook_on_press)
keyboard.on_release(keyboard_hook_on_release)


def load_image_bw(path):
    texture = Image.open(path)
    width, height = texture.size
    out = []
    for y in range(0, height):
        row = []
        for x in range(0, width):
            pixel = texture.getpixel((x, y))
            if pixel == (0, 0, 0, 255):
                row.append(True)
            else:
                row.append(False)
        out.append(row)
    return out


def load_image(path):
    return Image.open(path)


def set_background_collision(collision):
    global background_collision
    background_collision = collision


@drawCollision
def draw_background_collision():
    global background_collision

    if background_collision is not None:
        global game_map
        global engine_setting

        collision = background_collision.collision
        for k in range(0, smallest(len(collision[0]), len(game_map[0]))):
            for v in range(0, smallest((len(collision)), len(game_map))):
                if collision[v][k] is True:
                    game_map[v][k].add(background_collision)


def set_background(background):
    global current_background
    current_background = background


def set_resolution(width, height):
    global engine_setting
    engine_setting['resolution'] = (width, height)


def register(obj):
    global registry
    registry.add(obj)


def unregister(obj):
    global registry
    registry.remove(obj)


def distance(obj1, obj2):
    return sqrt(pow(obj2.x - obj1.x, 2) + pow(obj2.y - obj1.y, 2))


def render_tps_counter():
    global tps
    global engine_setting
    global game
    tps = str(int(tps))
    i = 0
    for l in tps:
        img = Image.open(os.path.dirname(os.path.abspath(__file__)) + '/assets/tps_counter/' + l + '.png')
        game.paste(img, (i, 0), img)
        i = i + 4


def engine_excepthook(type, value, _traceback):
    global engine_setting

    path = engine_setting['crash_path']

    if type != KeyboardInterrupt:
        with open(path, mode='w+') as file:
            file.write('Unhandled exception using Pyxel Engine ' + __version__ +
                       '\n\nTraceback (most recent call last):\n')
        with open(path, mode='a') as file:
            traceback.print_tb(_traceback, file=file)
        with open(path, mode='a') as file:
            file.write(str(type.__name__) + ': ' + str(value))

        logging.critical('Unhandled exception ' + str(type.__name__) + ': ' + str(value))

        sys.__excepthook__(type, value, _traceback)

        p = pathlib.Path(path)
        p.absolute()
        windll.user32.MessageBoxW(0, 'Crash report created in ' + str(p.absolute()), 'The game has crashed', 0)

        _thread.interrupt_main()
        exit()


def scale_to_screen(img):
    global screen_width
    global screen_height
    return img.resize((screen_width, screen_height))


def render():
    global window
    global game
    global engine_setting
    global current_background
    game.paste('#000000', [0, 0, game.size[0], game.size[1]])

    if current_background is not None:
        background = current_background.image.resize(engine_setting['resolution'], box=(current_background.y, current_background.x,
                                                                                        current_background.y + engine_setting['resolution'][0],
                                                                                        current_background.x + engine_setting['resolution'][1]))
        game.paste(background, (0, 0))

    logging.debug('Triggering renderTick and postProcessing')
    renderTick.trigger()
    postProcessing.trigger()

    window2 = scale_to_screen(game)
    window.paste(window2)


def main_loop():
    global tick
    global last_tick
    global next_tick
    global targetedTPS
    global tps

    logging.info('Engine launched')
    logging.debug('Triggering onEngineStarted')
    onEngineStarted.trigger()

    is_tps_counter_rendering = False

    while True:
        if engine_setting['TPS_counter'] is True and is_tps_counter_rendering is False:
            renderTick.add(render_tps_counter)
            is_tps_counter_rendering = True
        elif engine_setting['TPS_counter'] is False and is_tps_counter_rendering is True:
            renderTick.remove(render_tps_counter)
            is_tps_counter_rendering = False

        logging.debug('>>> TICK ' + str(tick))
        tick = tick + 1

        logging.debug('Triggering engineTick')
        engineTick.trigger()

        logging.debug('Started collision check')
        generate_game_map()
        drawCollision.trigger()
        checkCollision.trigger()

        logging.debug('Started frame rendering')
        render()
        logging.debug('Rendering finished')

        while time.time_ns() / (10 ** 9) < next_tick:
            pass

        tps = 1 / (time.time_ns() / (10 ** 9) - last_tick)

        last_tick = time.time_ns() / (10 ** 9)
        next_tick = time.time_ns() / (10 ** 9) + 1 / engine_setting['TPS']


def initialize_engine():
    global root
    global game
    global window
    global screen_width
    global screen_height
    global engine_setting

    logging.basicConfig(filename=engine_setting['log_path'], level=engine_setting['logging_level'],
                        format='[%(levelname)s] %(message)s')
    logging.info('Using Pyxel Engine ' + __version__)

    if engine_setting['catch_error'] is True:
        sys.excepthook = engine_excepthook

    generate_game_map()

    root = tk.Tk()
    screen_width = root.winfo_screenwidth()
    screen_height = root.winfo_screenheight()
    game = Image.new('RGB', engine_setting['resolution'], color='#000000')
    window = ImageTk.PhotoImage(image=scale_to_screen(game))

    logging.info('Engine initialized')


def generate_game_map():  # thanks to Ves Zappa#3787 (discord)
    global game_map

    game_map = []
    for i in range(engine_setting['resolution'][0]):
        blank_line = []
        for j in range(engine_setting['resolution'][1]):
            blank_line.append(set())
        game_map.append(blank_line)


def start_engine():
    if not engineInitialized:
        initialize_engine()
    global root
    global game

    root.attributes("-fullscreen", True)
    tk.Label(root, image=window).pack(fill=tk.BOTH, expand=1)
    _thread.start_new_thread(main_loop, ())
    root.mainloop()
